
from collections import defaultdict
import tensorflow as tf
import numpy as np
from time import time
from tqdm.notebook import tqdm


from WordEmbed.correlation import spearman_cor


# Building model

# Building layer

class Embed(tf.keras.layers.Layer):
    '''
    Define Embedding layer derived from keras layer. This layer will map words to ord embedding
    '''

    def __init__(self, vocab, units, name=None):
        super(Embed, self).__init__(name=name)
        self.units = units
        self.vocab = vocab

    def build(self, input_shape):
        w_init = tf.random_normal_initializer(seed=22)
        self.w = tf.Variable(initial_value=w_init(shape=[self.vocab, self.units], dtype=tf.float32),
                             trainable=True, name='kernel')

    def call(self, input):
        x = tf.one_hot(input, depth=self.vocab)
        return tf.matmul(x, self.w)


# Buiding Bias
class Bias(tf.keras.layers.Layer):
    '''
    Define Bias layer to train embeddings
    '''


    def __init__(self, vocab):
        super(Bias, self).__init__()
        self.vocab = vocab

    def build(self, input_shape):
        w_init = tf.random_normal_initializer(seed=20)
        self.w = tf.Variable(initial_value=w_init(shape=[self.vocab, 1], dtype=tf.float32),
                             trainable=True, name='kernel1')

    def call(self, input):
        x = tf.one_hot(input, depth=self.vocab)
        return tf.matmul(x, self.w)


# Define Model

class Model(tf.keras.Model):
    def __init__(self, vocab, embed_dim, Embed, Bias):
        super(Model, self).__init__()
        self.emb1 = Embed(vocab, embed_dim, name='embed')
        self.emb2 = Embed(vocab, embed_dim)
        self.b1 = Bias(vocab)
        self.b2 = Bias(vocab)
        self.dot = tf.keras.layers.Dot((1, 1))
        self.a = tf.keras.layers.Add()

    def call(self, inputs):
        i1 = inputs[0]
        i2 = inputs[1]
        o1 = self.emb1(i1)
        o2 = self.b1(i1)
        o3 = self.emb2(i2)
        o4 = self.b2(i2)
        o1 = self.dot([o1, o3])
        o = self.a([o1, o2, o4])
        return o


def train(model, d):
    opt = tf.keras.optimizers.Adam()
    x1, x2, y = d

    def data_gen(x1, x2, y):
        n = len(x1)
        for i in range(0, n, 1024):
            e = i + 1024
            if e < n:
                yield [x1[i:e], x2[i:e]], y[i:e]
            else:
                yield [x1[i:], x2[i:]], y[i:]

    # loss function
    @tf.function
    def loss_fn(a, y):
        a = tf.squeeze(a, axis=-1)
        # a=tf.cast(a,tf.float64)
        y = tf.convert_to_tensor(y)
        y = tf.cast(y, tf.float32)
        iszero = y <= 1.0
        f = tf.where(iszero, 0.25, 1.0)
        # f=tf.cast(f,tf.float64)
        # y=tf.cast(y,tf.float64)
        l = tf.math.square((a - tf.math.log(y + 0.1)))
        l = tf.multiply(f, l)
        loss = tf.reduce_sum(l)
        return loss

    @tf.function
    def train_step(input, y):
        with tf.GradientTape() as t:
            out = model(input)
            loss = loss_fn(out, y)
        var = model.trainable_variables
        grads = t.gradient(loss, var)
        opt.apply_gradients(zip(grads, var))
        return loss

    epochs = 20
    print('\nTraining starts...\n')
    for j in tqdm(range(epochs)):
        #print('\n\nEpochs : {}'.format(j + 1))
        t = time()
        l = 0
        i = 0
        l_ = []
        for inp, y1 in data_gen(x1, x2, y):
            ls = train_step(inp, y1)
            l = l + ls
            # print(ls)
            i = i + 1
        l = l / i
        #print('Loss: {}\nTime: {}'.format(l, time() - t))
        l_.append(l)




class Glove():
    def __init__(self,
                 data,
                 embedding_dim=None,
                 window=2,
                 trainable=True,
                 ):
        self.window = window
        self.trainable = trainable
        self.model = None
        self.data = data
        self.vocab_size = None
        self.embedding_dim = embedding_dim
        self.embeddings = {}
        self.tokenizer = None
        self.score = None
        self.TfAPI = None

    def process_data(self, data):

        # Tokenize the data
        tokenizer = tf.keras.preprocessing.text.Tokenizer(num_words=None,
                                                          oov_token='<oov>',
                                                          filters='!"#$%&()*+.,-/:;=?@[\]^_`{|}~ ')
        tokenizer.fit_on_texts(data)
        proc_data = tokenizer.texts_to_sequences(data)

        self.vocab_size = len(tokenizer.index_word) + 1
        self.tokenizer = tokenizer
        return proc_data

    # Defining matrix
    def build_mtrx(self, sentences, vocab, win):
        n = len(vocab) + 1
        mtrx = defaultdict(lambda: defaultdict(int))
        print('\ncreating matrix...\n')
        for i in tqdm(sentences):
            for j in range(len(i)):
                for d in range(1, win + 1):
                    if (j + d) < len(i):
                        w = 1 / d
                        x, y = sorted([i[j], i[j + d]])
                        mtrx[x][y] = mtrx[x][y] + w
        return mtrx

    def extr(self, dic):
        x = []
        y = []
        w = []
        for i in dic.keys():
            for j in dic[i].keys():
                x.append(i)
                y.append(j)
                w.append(dic[i][j])
        return np.array(x), np.array(y), np.array(w)

    def train(self):
        # Process the data
        data = self.process_data(self.data)

        # Build matrix
        vocab = self.tokenizer.word_index
        mtrx = self.build_mtrx(data, vocab, self.window)
        x1, x2, y = self.extr(mtrx)

        # Define the model
        self.model = Model(self.vocab_size, self.embedding_dim, Embed, Bias)

        # Train the model
        train(self.model, [x1, x2, y])

        # Save the embeddings
        weights = self.model.get_layer(name='embed').get_weights()[0]
        layer = self.model.get_layer(name='embed')
        vocab = self.tokenizer.index_word
        for i in range(1, len(weights)):
            self.embeddings[vocab[i]] = weights[i]

        # Calculating score/quality of embeddings
        self.score = spearman_cor(self.embeddings)

        # forming TF API
        if self.trainable == False:
            layer.trainable = False
        self.TfAPI = layer


'''
""" Test the class"""

# Loading shakespear data for our analysis
path_to_file = tf.keras.utils.get_file('shakespeare.txt',
                                       'https://storage.googleapis.com/download.tensorflow.org/data/shakespeare.txt')

with open(path_to_file) as f:
    data = f.read().splitlines()

emb = Glove(data, 512)

emb.train()

'''




import tensorflow as tf
import numpy as np
import random
from tqdm.notebook import tqdm
import tensorflow.keras.backend as K

from WordEmbed.correlation import spearman_cor


# create the model
class Model(tf.keras.Model):
    def __init__(self, vocab_size, embed_dim, neg_sample):
        super(Model, self).__init__()
        self.target_embed = tf.keras.layers.Embedding(vocab_size,
                                                      embed_dim,
                                                      input_length=1,
                                                      name='embed')
        self.context_embed = tf.keras.layers.Embedding(vocab_size,
                                                       embed_dim,
                                                       input_length=neg_sample)

        self.lambd = tf.keras.layers.Lambda(lambda x: K.mean(x, axis=1))
        self.dot = tf.keras.layers.Dot((1, 2))

    def call(self, inputs):
        t = inputs[0]
        c = inputs[1]
        x = self.context_embed(t)
        y = self.target_embed(c)
        x = self.lambd(x)
        z = self.dot([x, y])
        out = tf.keras.activations.softmax(z)
        return out


class CBOW():
    def __init__(self,
                 data,
                 embedding_dim=None,
                 sampling_window=2,
                 negative_samples=4,
                 trainable=True,
                 ):
        self.trainable = trainable
        self.model = None
        self.sampling_window = sampling_window
        self.negative_samples = negative_samples
        self.data = data
        self.vocab_size = None
        self.embedding_dim = embedding_dim
        self.embeddings = {}
        self.tokenizer = None
        self.score = None
        self.TfAPI = None

    def process_data(self, data):

        # To create training data minimum sentence length=window+neg_sample
        n = self.negative_samples + self.sampling_window
        data = [i for i in data if len(i.split()) >= n]

        # Tokenize the data
        tokenizer = tf.keras.preprocessing.text.Tokenizer(num_words=None,
                                                          oov_token='<oov>',
                                                          filters='!"#$%&()*+.,-/:;=?@[\]^_`{|}~ ')
        tokenizer.fit_on_texts(data)
        proc_data = tokenizer.texts_to_sequences(data)

        self.vocab_size = len(tokenizer.index_word) + 1
        self.tokenizer = tokenizer
        return proc_data

    def data_sampling(self, data, win, n):
        context = []
        label = []
        target = []
        print('\nSampling the data..\n')
        for a in tqdm(data):
            for i in range(len(a)):
                if len(a) > (win + n + 1):
                    if i == 0:
                        con = a[1:win + 1]
                        n_sample = a[win + 1:]
                        ns = random.sample(n_sample, n)
                        target.append([a[i]] + ns)
                        context.append(con)
                        label.append([1] + [0] * n)
                    elif i == len(a):
                        con = a[n - win - 1:-1]
                        n_samle = a[:len(a) - win - 1]
                        ns = random.sample(n_sample, n)
                        target.append([a[i]] + ns)
                        context.append(con)
                        label.append([1] + [0] * n)
                    else:
                        st = max(0, i - win)
                        ed = min(len(a), i + win + 1)
                        con = a[st:i] + a[i + 1:ed]
                        n_sample = a[:st] + a[ed:]
                        if len(n_sample) > (n - 1):
                            ns = random.sample(n_sample, n)
                            target.append([a[i]] + ns)
                            context.append(con)
                            label.append([1] + [0] * n)
                        elif len(a) > 6:
                            context.append([a[i - 1], a[i + 1]])
                            n_sample = a[:i - 1] + a[i + 2:]
                            ns = random.sample(n_sample, n)
                            target.append([a[i]] + ns)
                            label.append([1] + [0] * n)
        return context, target, label

    def train(self):
        # Process the data
        data = self.process_data(self.data)

        # Define the model
        self.model = Model(self.vocab_size, self.embedding_dim, self.negative_samples)

        # Get the data for training
        context, target, label = self.data_sampling(data, self.sampling_window, self.negative_samples)
        a = np.array([context, target, label], dtype=object)
        a3 = np.array([[a[0][i], a[1][i], a[2][i]] for i in range(len(a[0])) if len(a[0][i]) == self.negative_samples],
                      dtype=object)
        con = list(a3[:, 0])
        tar = list(a3[:, 1])
        lab = list(a3[:, 2])

        trn = tf.data.Dataset.from_tensor_slices(((con, tar), lab))
        trn = trn.batch(128).shuffle(buffer_size=20000).prefetch(buffer_size=tf.data.AUTOTUNE)

        # Train the model
        print('\nTraining starts ...')
        self.model.compile(optimizer='adam', loss="categorical_crossentropy", metrics=['accuracy'])
        self.model.fit(trn, epochs=10, verbose=1)

        # Save the embeddings
        weights = self.model.get_layer(name='embed').get_weights()[0]
        layer = self.model.get_layer(name='embed')
        vocab = self.tokenizer.index_word
        for i in range(1, len(weights)):
            self.embeddings[vocab[i]] = weights[i]

        # Calculating score/quality of embeddings
        self.score = spearman_cor(self.embeddings)

        # forming TF API
        if self.trainable == False:
            layer.trainable = False
        self.TfAPI = layer



'''
"""# Test the class"""

# Loading shakespear data for our analysis
path_to_file = tf.keras.utils.get_file('shakespeare.txt',
                                       'https://storage.googleapis.com/download.tensorflow.org/data/shakespeare.txt')

with open(path_to_file) as f:
    data = f.read().splitlines()

w2v = CBOW(data, 512)

w2v.train()

print(w2v.score)

print(w2v.TfAPI)

# w2v.embeddings

'''


import pandas as pd
import numpy as np
import requests, zipfile, io

def spearman_cor(embed):
    # Download Simlex file and save it locally
    r = requests.get('https://fh295.github.io/SimLex-999.zip', stream=True)
    z = zipfile.ZipFile(io.BytesIO(r.content))
    z.extractall()

    # Load simlex words
    df = pd.read_csv('SimLex-999/SimLex-999.txt',sep='\t')
    word1 = list(df['word1'])
    word2 = list(df['word2'])
    simlex_words = set(word1 + word2)

    # Load trained embedding
    embed_words = list(embed.keys())
    embed_words = set(embed_words)

    # common words
    common_words = embed_words.intersection(simlex_words)

    # Compress simplex to only common words
    df1 = df.loc[(df['word1'].isin(common_words)) & (df['word2'].isin(common_words))]
    word1 = list(df1['word1'])
    word2 = list(df1['word2'])
    sim_index = list(df1['SimLex999'])

    # Calculate Cosine Simalirity
    cos_sim = []
    for a, b in zip(word1, word2):
        cos_sim.append(np.dot(embed[a], embed[b]) / (np.linalg.norm(embed[a]) * np.linalg.norm(embed[b])))

    # Calculate difference in ranks
    sim_index = -1 * np.array(sim_index)
    cos_sim = -1 * np.array(cos_sim)
    emb_rank = np.argsort(cos_sim)
    sim_rank = np.argsort(sim_index)
    distance = abs(emb_rank - sim_rank)

    # Calculate spearman correlation
    n = len(distance)
    rho = 1 - 6 * np.sum(np.square(distance)) / (n * (n ** 2 - 1))
    return rho



import tensorflow as tf
import random
from tqdm.notebook import tqdm


from WordEmbed.correlation import spearman_cor


class Model(tf.keras.Model):
    def __init__(self, vocab_size, embed_dim, neg_sample):
        super(Model, self).__init__()
        self.target_embed = tf.keras.layers.Embedding(vocab_size,
                                                      embed_dim,
                                                      input_length=1,
                                                      name='embed')
        self.context_embed = tf.keras.layers.Embedding(vocab_size,
                                                       embed_dim,
                                                       input_length=neg_sample)
        self.dot = z = tf.keras.layers.Dot((1, 2))

    @tf.function
    def call(self, inputs):
        i1 = inputs[0]
        i2 = inputs[1]
        x = self.target_embed(i1)
        y = self.context_embed(i2)
        x = tf.squeeze(x, axis=[1])
        z = self.dot([x, y])
        out = tf.keras.activations.softmax(z)
        return out




class Skipgram():
    def __init__(self,
                 data,
                 embedding_dim=None,
                 sampling_window=2,
                 negative_samples=4,
                 trainable=True,
                 ):
        self.trainable = trainable
        self.model = None
        self.sampling_window = sampling_window
        self.negative_samples = negative_samples
        self.data = data
        self.vocab_size = None
        self.embedding_dim = embedding_dim
        self.embeddings = {}
        self.tokenizer = None
        self.score = None
        self.TfAPI = None

    def process_data(self, data):

        # To create training data minimum sentence length=window+neg_sample
        n = self.negative_samples + self.sampling_window
        data = [i for i in data if len(i.split()) >= n]

        # Tokenize the data
        tokenizer = tf.keras.preprocessing.text.Tokenizer(num_words=None,
                                                          oov_token='<oov>',
                                                          filters='!"#$%&()*+.,-/:;=?@[\]^_`{|}~ ')
        tokenizer.fit_on_texts(data)
        proc_data = tokenizer.texts_to_sequences(data)

        self.vocab_size = len(tokenizer.index_word) + 1
        self.tokenizer = tokenizer
        return proc_data

    def data_sampling(self, sequences, n_sampl, win):
        lab, tar, con = [], [], []
        print('\nSampling the data..\n')
        for seq in tqdm(sequences):
            if len(seq) > (win + n_sampl + 1):
                tar.append(seq[0])
                con.append([random.choice(seq[1:1 + win])] + random.sample(seq[win + 2:], n_sampl))
                lab.append([1] + [0] * n_sampl)
                tar.append(seq[-1])
                con.append([random.choice(seq[-(win + 1):-1])] + random.sample(seq[:-(win + 1)], n_sampl))
                lab.append([1] + [0] * n_sampl)
                for i in range(1, len(seq) - 1):
                    st = max(0, i - win)
                    ed = min(len(seq) - 1, i + win + 1)
                    c = seq[st:i] + seq[i + 1:ed]
                    ns = seq[:st] + seq[ed:]
                    if len(ns) > (n_sampl - 1):
                        con_ar = [random.choice(c)] + random.sample(ns, n_sampl)
                        tar.append(seq[i])
                        con.append(con_ar)
                        lab.append([1] + [0] * n_sampl)
        return tar, con, lab

    def train(self):
        # Process the data
        data = self.process_data(self.data)

        # Define the model
        self.model = Model(self.vocab_size, self.embedding_dim, self.negative_samples)

        # Get the data for training
        target, context, label = self.data_sampling(data, self.negative_samples, self.sampling_window)
        trn = tf.data.Dataset.from_tensor_slices(((target, context), label))
        trn = trn.batch(128).shuffle(buffer_size=20000).prefetch(buffer_size=tf.data.AUTOTUNE)

        # Train the model
        self.model.compile(optimizer='adam', loss="categorical_crossentropy", metrics=['accuracy'])
        self.model.fit(trn, epochs=1)

        # Save the embeddings
        weights = self.model.get_layer(name='embed').get_weights()[0]
        layer = self.model.get_layer(name='embed')
        vocab = self.tokenizer.index_word
        for i in range(1, len(weights)):
            self.embeddings[vocab[i]] = weights[i]

        # Calculating score/quality of embeddings
        self.score = spearman_cor(self.embeddings)

        # forming TF API
        if self.trainable == False:
            layer.trainable = False
        self.TfAPI = layer




'''

"""#Test the class"""

# Loading shakespear data for our analysis
path_to_file = tf.keras.utils.get_file('shakespeare.txt',
                                       'https://storage.googleapis.com/download.tensorflow.org/data/shakespeare.txt')

with open(path_to_file) as f:
    data = f.read().splitlines()

w2v = Skipgram(data, 512)

w2v.train()

print(w2v.score)

l = w2v.TfAPI

'''




from WordEmbed.skipgram import Skipgram
from WordEmbed.cbow import CBOW
from WordEmbed.glove import Glove


def Embed(
        data,
        embedding_dim,
        embedding='glove',
        sampling_window=2,
        negative_samples=4,
        Trainable=True ):

    emb={'glove':Glove,'skipgram':Skipgram,'cbow':CBOW}
    if embedding not in emb.keys():
        raise Exception('embedding type is not from listed options')
    if sampling_window<1 and sampling_window>10:
        raise Exception('Sampling window is not in expected range')
    if negative_samples<2 and negative_samples>15:
        raise Exception('Negative samples are not in range')

    if embedding=='glove':
        return emb[embedding](data,embedding_dim,sampling_window,Trainable)
    if embedding=='skipgram':
        return emb[embedding](data,embedding_dim,sampling_window,negative_samples,Trainable)
    if embedding== 'cbow':
        return emb[embedding](data,embedding_dim,sampling_window,trainable)



